import os
import sys
import unittest
from unittest.mock import patch

from terrasafe import terrasafe


class TerrasafeTest(unittest.TestCase):
    resource_delete = {
        "change": {
            "actions": [
                "delete"
            ]
        }
    }

    resource_recreate = {
        "change": {
            "actions": [
                "delete",
                "create"
            ]
        }
    }

    resource_update = {
        "change": {
            "actions": [
                "update"
            ]
        }
    }

    def setUp(self):
        self.input_file = open('test/test-input.json', 'r')
        sys.stdin = self.input_file

    def tearDown(self):
        self.input_file.close()

    def test_load_config(self):
        config = terrasafe.load_config("./test/test-config.json")
        self.assertListEqual(config["ignore_deletion"],
                             ["aws_ecs_task_definition*", "1"])
        self.assertListEqual(config["ignore_deletion_if_recreation"],
                             ["aws_ecs_task_definition*", "2"])
        self.assertListEqual(config["unauthorized_deletion"],
                             ["aws_ecs_task_definition*", "3"])

    def test_parse_ignored_from_env_var(self):
        os.environ["TERRASAFE_ALLOW_DELETION"] = "a;b;c"
        res = terrasafe.parse_ignored_from_env_var()
        self.assertListEqual(res, ["a", "b", "c"])

    def test_parse_ignored_from_empty_env_var(self):
        os.environ["TERRASAFE_ALLOW_DELETION"] = ""
        res = terrasafe.parse_ignored_from_env_var()
        self.assertListEqual(res, [])

    def test_get_resource_deletion(self):
        all_deletion = terrasafe.get_resource_deletion()
        all_deletion_address = list(map(lambda res: res["address"], all_deletion))
        self.assertListEqual(all_deletion_address, ['aws_s3_bucket.b[0]', 'aws_test.test'])

    def test_has_delete_action(self):
        self.assertTrue(terrasafe.has_delete_action(self.resource_delete))
        self.assertTrue(terrasafe.has_delete_action(self.resource_recreate))
        self.assertFalse(terrasafe.has_delete_action(self.resource_update))

    def test_is_resource_recreate(self):
        self.assertFalse(terrasafe.is_resource_recreate(self.resource_delete))
        self.assertTrue(terrasafe.is_resource_recreate(self.resource_recreate))
        self.assertFalse(terrasafe.is_resource_recreate(self.resource_update))

    def test_is_resource_match_any(self):
        self.assertTrue(terrasafe.is_resource_match_any("ab.cd", ["c", "d", "ab*"]))
        self.assertFalse(terrasafe.is_resource_match_any("ab.cd", ["c", "ab"]))

    def test_is_deletion_commented(self):
        self.assertFalse(terrasafe.is_deletion_commented("test", "test"))
        self.assertFalse(terrasafe.is_deletion_commented("aws_s3_bucket", "b"))
        self.assertFalse(terrasafe.is_deletion_commented("aws_s3_bucket", "other_bucket"))
        self.assertTrue(terrasafe.is_deletion_commented("aws_type", "abc"))
        self.assertTrue(terrasafe.is_deletion_commented("aws_type2", "def"))

    def test_is_deletion_in_disabled_file(self):
        self.assertTrue(terrasafe.is_deletion_in_disabled_file("aws_type3", "disabled"))
        self.assertFalse(terrasafe.is_deletion_in_disabled_file("aws_s3_bucket", "b"))

    def test_all_tf_files(self):
        self.assertCountEqual(terrasafe.get_all_files(".tf"), ["./test/test.tf", "./test/test2.tf"])

    def test_main_fail(self):
        with patch.object(sys, 'argv', ["terrasafe"]), self.assertRaises(SystemExit):
            terrasafe.main()

    def test_main_fail_with_unauthorized(self):
        with patch.object(sys, 'argv', ["terrasafe", "--config", "./test/test-config-unauthorized.json"]), \
             self.assertRaises(SystemExit):
            terrasafe.main()

    def test_main_ok_ignore(self):
        with patch.object(sys, 'argv', ["terrasafe", "--config", "./test/test-config-ignore-all.json"]):
            terrasafe.main()

    def test_main_ok_ignore_with_var(self):
        os.environ["TERRASAFE_ALLOW_DELETION"] = "aws_s3_bucket.b[0];aws_test.*"
        with patch.object(sys, 'argv', ["terrasafe"]):
            terrasafe.main()



#!/usr/bin/python3
import os
import fnmatch
import re
from argparse import ArgumentParser
import json
import sys


def main():
    config = {"ignore_deletion": [], "ignore_deletion_if_recreation": [], "unauthorized_deletion": []}

    args = parse_argument()
    if args.config:
        config = {**config, **load_config(args.config)}
    else:
        print("No config provided. Default values used.")

    ignored_from_env_var = parse_ignored_from_env_var()

    all_deletion = get_resource_deletion()

    # List of all deletion which is not whitelisted or commented. Fail if not empty
    unauthorized_deletion = []

    for resource in all_deletion:
        resource_address = resource["address"]
        if is_resource_match_any(resource_address, config["unauthorized_deletion"]):
            print(f'Resource {resource_address} can not be destroyed for any reason')
            exit(1)
        if is_resource_match_any(resource_address, config["ignore_deletion"]):
            continue
        if is_resource_recreate(resource) and is_resource_match_any(resource_address,
                                                                    config["ignore_deletion_if_recreation"]):
            continue
        if is_resource_match_any(resource_address, ignored_from_env_var):
            print(f"deletion of {resource_address} authorized by env var.")
            continue
        if is_deletion_commented(resource["type"], resource["name"]):
            print(f"deletion of {resource_address} authorized by comment")
            continue
        if is_deletion_in_disabled_file(resource["type"], resource["name"]):
            print(f"deletion of {resource_address} authorized by disabled file feature")
            continue
        unauthorized_deletion.append(resource_address)

    if unauthorized_deletion:
        print("Unauthorized deletion detected for those resources:")
        for deletion in unauthorized_deletion:
            print(f" - {deletion}")
        print("If you really want to delete those resources: comment it or export this environment variable:")
        print(f"export TERRASAFE_ALLOW_DELETION=\"{';'.join(unauthorized_deletion)}\"")
        exit(1)
    else:
        print("0 unauthorized deletion detected")


def parse_argument():
    args_parser = ArgumentParser()
    args_parser.add_argument("--config",
                             help="specify the path to the terrasafe config",
                             required=False
                             )
    return args_parser.parse_args()


def load_config(config_path):
    json_file = open(config_path)
    config = json.load(json_file)
    print("Config loaded from", config_path)
    json_file.close()
    return config


def parse_ignored_from_env_var():
    ignored = os.environ.get("TERRASAFE_ALLOW_DELETION")
    if ignored:
        return ignored.split(";")
    return []


def get_resource_deletion():
    data = json.load(sys.stdin)
    # check format version
    if data["format_version"].split(".")[0] != "0":
        print("Only format major version 0 is supported")
        exit(1)
    if "resource_changes" in data:
        resource_changes = data["resource_changes"]
    else:
        resource_changes = []

    return list(filter(has_delete_action, resource_changes))


def has_delete_action(resource):
    return "delete" in resource["change"]["actions"]


def is_resource_match_any(resource_address, pattern_list):
    for pattern in pattern_list:
        pattern = re.sub(r"\[(.+?)\]", "[[]\g<1>[]]", pattern)
        if fnmatch.fnmatch(resource_address, pattern):
            return True
    return False


def is_resource_recreate(resource):
    actions = resource["change"]["actions"]
    return "create" in actions and "delete" in actions


def is_deletion_commented(resource_type, resource_name):
    regex = re.compile(rf'(#|//)\s*resource\s*\"{resource_type}\"\s*\"{resource_name}\"')
    tf_files = get_all_files(".tf")
    for filepath in tf_files:
        with open(filepath, 'r') as file:
            for line in file:
                if regex.match(line):
                    return True
    return False


def is_deletion_in_disabled_file(resource_type, resource_name):
    regex = re.compile(rf'\s*resource\s*\"{resource_type}\"\s*\"{resource_name}\"')
    tf_files = get_all_files(".tf.disabled")
    for filepath in tf_files:
        with open(filepath, 'r') as file:
            for line in file:
                if regex.match(line):
                    return True


def get_all_files(extension):
    res = []
    for root, dirs, file_names in os.walk("."):
        for file_name in file_names:
            if fnmatch.fnmatch(file_name, "*" + extension):
                res.append(os.path.join(root, file_name))
    return res


if __name__ == '__main__':
    main()



from setuptools import setup, find_packages
import os

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='terrasafe',
    version=os.getenv('RELEASE_TAG'),
    packages=find_packages(".", exclude=["test"]),
    url='https://github.com/PrismeaOpsTeam/Terrasafe',
    license='GPLv3',
    author='PrismeaOpsTeam',
    author_email='team-ops@prismea.fr',
    description='A CLI tool to analyze Terraform plan files, and prevent unallowed resources deletions and drop/create actions.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    python_requires='>=3.8',
    entry_points={
        'console_scripts': [
            'terrasafe = terrasafe.terrasafe:main'
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],

)
